# -*- coding: utf-8 -*-
# Copyright 2009-2015, Red Hat, Inc.
# License: GPL-2.0+ <http://spdx.org/licenses/GPL-2.0+>
# See the LICENSE file for more details on Licensing


"""Interface to locally spawned virtual machines that are used as disposable
clients for executing Taskotron tasks."""

from time import sleep, time
import socket
import getpass
import re
import os

from libtaskotron.logger import log
from libtaskotron import config
import libtaskotron.exceptions as exc

from testcloud import instance, image, cli
from testcloud.exceptions import TestcloudInstanceError, TestcloudImageError


class TestCloudMachine(object):
    '''Launch virtual machines with TestCloud and prepare them for executing
    tasks (install packages etc.)'''

    def __init__(self, uuid):
        ''':param uuid: unicode string uuid for the task being executed'''

        # this should be temporary and go away once we have better image
        # finding and discovery.
        self.imageurl = config.get_config().imageurl

        #: name of the testcloud instance spawned for this task
        self.instancename = 'taskotron-{}'.format(uuid)

        #: username to use when connecting to the virtual machine
        self.username = 'root'

        #: password for user on the remote machine
        self.password = 'passw0rd'

        #: ip address of the remote machine (only set after prepared and spawned successfully)
        self.ipaddr = None

        #: uuid of the task which spawned this instance
        self.uuid = uuid

        #: hostname to use for spawned instance - based on username of current user
        self.hostname = 'taskotron-%s' % getpass.getuser()

    def _prepare_image(self):
        '''Use testcloud to prepare an image for local booting
        :raises TaskotronRemoteError: for errors in preparing the image with testcloud'''

        log.debug("preparing image {} for task {}".format(self.imageurl, self.uuid))
        tc_image = None

        try:
            tc_image = image.Image(self.imageurl)
            tc_image.prepare()
        except TestcloudImageError, e:
            log.error(e)
            raise exc.TaskotronRemoteError("There was an error while preparing the "
                                           "testcloud image", e)

        return tc_image

    def _prepare_instance(self, tc_image):
        '''Prepare an instance for booting and boot it with testcloud'''
        log.debug("preparing testcloud instance {}".format(self.instancename))
        tc_instance = instance.Instance(self.instancename, tc_image, hostname=self.hostname)
        tc_instance.prepare()

        log.debug("spawning testcloud instance {}".format(self.instancename))
        tc_instance.spawn_vm()
        tc_instance.start()

    def _check_existing_instance(self, should_exist=False):
        '''Check whether an instance with the same name already exists, raise errors if the result
        was not as expected.

        :param should_exist: sets expectation on whether the instance should
                             exist already or not
        :raises TaskotronRemoteError: if the preset expectation is not met
        '''
        existing_instance = instance.find_instance(self.instancename)

        if existing_instance is None:
            if should_exist:
                raise exc.TaskotronRemoteError("Was expecting to find instance {} but it does not"
                                               " already exist".format(self.instancename))
        else:
            if not should_exist:
                raise exc.TaskotronRemoteError("Expected to NOT find instance {} but it is "
                                               "already defined".format(self.instancename))
        return existing_instance

    def prepare(self):
        '''Prepare a virtual machine for running tasks.

        :raises TaskotronRemoteError: if there are any errors while preparing
                                      the virtual machine.
       '''

        tc_image = self._prepare_image()

        self._check_existing_instance(should_exist=False)

        self._prepare_instance(tc_image)

        self.ipaddr = cli.find_vm_ip(self.instancename)

    def wait_for_port(self, port, timeout=60):
        '''Wait until port is open. Repeatedly tries to socket.connect on given port.

        :param port: port to check
        :param timeout: timeout in seconds
        :raises TaskotronInstanceError: when timeouted
        '''
        s = socket.socket()
        start_time = time()
        log.debug('Waiting up to %d seconds for %s:%s to open.' % (timeout, self.ipaddr, port))
        while True:
            try:
                s.connect((self.ipaddr, port))
            except socket.error:
                pass
            else:
                s.close()
                break
            if (start_time + timeout) < time():
                raise TestcloudInstanceError("Waiting for %s:%s to open timed out (%ds)" %
                                             (self.ipaddr, port, timeout))

            sleep(0.1)

    def _stop_instance(self, tc_instance):
        '''Stop (shut down) a virtual machine instance

        :raises TaskotronRemoteError: if stopping the instance fails'''

        log.debug("Stopping instance {}".format(self.instancename))
        try:
            tc_instance.stop()
        except TestcloudInstanceError, e:
            log.error("Hit error while stopping instance {}".format(self.instancename), e)
            raise exc.TaskotronRemoteError(e)

    def _destroy_instance(self, tc_instance):
        '''Destroy (delete) the virtual machine instance.

        :raises TaskotronRemoteError: if destroying the instance fails'''

        log.debug("destroying instance {}".format(self.instancename))
        try:
            tc_instance.destroy()
        except TestcloudInstanceError, e:
            log.error("Hit error while destroying instance {}".format(self.instancename), e)
            raise exc.TaskotronRemoteError(e)

    def teardown(self):
        '''Tear down the virtual machine by shutting it down and removing any
        trace of it from the host machine.

        :raises TaskotronRemoteError: if there is a failure while stopping or
                                      removing the virtual machine instance
        '''

        tc_instance = self._check_existing_instance(should_exist=True)

        self._stop_instance(tc_instance)

        # wait for testcloud to actually stop the instance, currently
        # instance.stop() returns before the instance is actually stopped
        sleep(2)

        self._destroy_instance(tc_instance)


class Image(object):
    '''Retrieve images from either local or remote sources'''

    @classmethod
    def get_all_filenames(cls, imagesdir=None):
        """Get list of images present on the system.

        :param imagesdir: absolute path to directory containing the images, path from config is
                          used if None
        """

        if not imagesdir:
            imagesdir = config.get_config().imagesdir

        return os.listdir(imagesdir)

    @classmethod
    def get_latest(cls, release, flavor, arch='x86_64', imagesdir=None):
        """Search for the most recent image available on the system.

        :param release: release of the image (e.g. ``23``)
        :param flavor: flavor of the image (e.g. ``minimal``)
        :param imagesdir: absolute path to directory containing the images, path from config is
                          used if None
        :param arch: architecture of the image

        :return: testcloud.image.Image instance
        :raises TaskotronImageError: if no such image for given release and flavor was found
        """

        latest_metadata = cls.get_latest_metadata(release, flavor, arch, imagesdir)

        if not latest_metadata:
            raise exc.TaskotronImageNotFoundError('No such image for release %s and flavor %s.' %
                                                  (release, flavor))
        else:
            return image.Image("file://" + os.path.join(imagesdir, latest_metadata['filename']))

    @classmethod
    def get_latest_metadata(cls, release, flavor, arch='x86_64', imagesdir=None):
        """Search for the most recent image available on the system.

        :param release: release of the image (e.g. ``23``)
        :param flavor: flavor of the image (e.g. ``minimal``)
        :param imagesdir: absolute path to directory containing the images, path from config is
                          used if None
        :param arch: arch of the image (e.g. 'x86_64')

        :return: metadata of the most recent image
        :rtype: dict {'date': str, 'version': str, 'release': str, 'arch': str, 'filename': str}
        """

        pattern = re.compile(r'^([0-9]{8})_([0-9]{4})-fedora-(.*?)-(.*?)-(.*?)\.qcow2$')

        images = []
        for filename in cls.get_all_filenames(imagesdir):
            m = pattern.match(filename)
            if m:
                images.append({'date': m.group(1),
                               'version': m.group(2),
                               'release': m.group(3),
                               'flavor': m.group(4),
                               'arch': m.group(5),
                               'filename': filename})

        filtered = [i for i in images if i['release'] == release and i['flavor'] == flavor and i['arch'] == arch]
        if not filtered:
            return None
        else:
            return sorted(filtered, key=lambda i: int(i['date']+i['version']))[-1]
